import re
from pathlib import Path
from typing import List, Optional

import click

from kks.util.common import find_workspace, find_problem_rootdir, config_directory


target_file = 'targets.yaml'

global_comment = '# This is the default config file, it is used in any subdirectory of the workspace.\n'\
                 '# You can modify the default target (the changes will be applied only in this workspace).\n'\
                 '# Also, you can add/overwrite targets by creating another "targets.yaml" file in your working directory.\n'\
                 '\n'

class Target:
    class Options:
        """A class containing type hints for all parsed options."""
        compiler: Optional[str]
        cpp_compiler: Optional[str]
        std: Optional[str]
        cpp_std: Optional[str]
        flags: Optional[List[str]]
        files: Optional[List[str]]
        libs: Optional[List[str]]
        asm64bit: Optional[bool]
        default_asan: Optional[bool]
        out: Optional[str]

        @classmethod
        def names(cls):
            return cls.__annotations__.keys()

    def __init__(self, name, settings):
        self.name = name
        self.need_default = False
        self.parent = None

        for opt_name in self.Options.names():
            value = settings.get(opt_name)
            setattr(self, opt_name, value)
            # option is not set or the list contains "DEFAULT" as first item
            # if the list is empty, we shouldn't replace it (e.g. if this target excludes all libs)
            if value is None or (isinstance(value, list) and value and value[0] == 'DEFAULT'):
                self.need_default = True

    def __str__(self):
        options = [
            f'{opt_name}={getattr(self, opt_name)}' for opt_name in self.Options.names()
        ]
        return f'Target("{self.name}", {", ".join(options)}")'

    def set_parent(self, parent: Optional['Target']):
        """Insert a target into the inheritance chain as the parent of this target."""
        if parent is not None:
            parent.parent = self.parent
        self.parent = parent

    def resolve_options(self, problem):
        """Replace macros and pull missing options from parents."""
        if self.need_default:
            assert self.parent is not None
            self.parent.resolve_options(problem)

        def modify(x):
            return x.replace('TASKNAME', problem)

        def modify_list(lst):
            if len(lst) == 0:
                return lst
            if lst[0] == 'DEFAULT':
                assert self.parent is not None
                default = getattr(self.parent, opt_name)
                lst = default + lst[1:]
            return [modify(e) for e in lst]

        for opt_name in self.Options.names():
            opt = getattr(self, opt_name)
            if opt is None:
                assert self.parent is not None
                opt = getattr(self.parent, opt_name)
            if isinstance(opt, list):
                opt = modify_list(opt)
            elif isinstance(opt, str):
                opt = modify(opt)
            setattr(self, opt_name, opt)


def check_version(cfg_file, cfg, new_version, is_global=False):
    if cfg.get('__version__', new_version) == new_version:
        return

    old_version = cfg['__version__']
    click.echo(click.style(str(cfg_file.absolute()), fg='blue', bold=True) +
               click.style(' is outdated. You can run "kks init --config=update" if you want to update the default target manually (see README on Github)', fg='yellow'))
    cfg_file.write_text(re.sub(rf'^(__version__: +){old_version}', rf'\g<1>{new_version}', cfg_file.read_text(), 1, re.M))


def find_target(name):
    import yaml

    def load_config(config_file):
        with config_file.open('r') as f:
            try:
                config = yaml.safe_load(f)
                return config
            except yaml.parser.ParserError:
                click.secho(f'Cannot parse {config_file.absolute()}', fg='red', err=True)
                return None

    def get_target(config, target_name):
        if config is None:
            return None
        if target_name in config:
            return Target(target_name, config[target_name])
        return None

    # we don't use pkg_resources, because it takes 300+ ms to load
    package_cfg = load_config(Path(__file__).parents[1] / 'data' / target_file)

    # package_default should have all fields set
    package_default = Target('default', package_cfg['default'])

    cwd = Path.cwd()
    workspace = find_workspace()
    rootdir = find_problem_rootdir()
    problem = '{}-{}'.format(*rootdir.parts[-2:]) if rootdir is not None else 'sm00-0'

    if workspace is not None:
        root_file = workspace / target_file  # we load root config even if we won't need it
        root_cfg = load_config(root_file) if root_file.is_file() else None
        if root_cfg is not None:
            check_version(root_file, root_cfg, package_cfg['__version__'], True)
    else:
        root_cfg = None

    if workspace is None or cwd != workspace:
        local_file = cwd / target_file
        local_cfg = load_config(local_file) if local_file.is_file() else None
        if local_cfg is not None:
            check_version(local_file, local_cfg, package_cfg['__version__'])
    else:
        local_cfg = None

    # Any target except package_default may have missing fields (e.g. an otdated custom target file).
    # In this case we need to inherit from package_default.
    # As a side effect, custom default targets are able to use the DEFAULT macro.
    # The result of using DEFAULT in a default target is undefined.

    # Inheritance chain:
    # local_default -> [[root_default] -> package_default] -> None
    # local_non_default -> [[local_default or root_default] -> package_default] -> None
    local_target = get_target(local_cfg, name)
    if local_target is not None:
        if local_target.need_default:
            local_target.set_parent(package_default)
            if name == 'default':
                custom_default = get_target(root_cfg, 'default')
            else:
                custom_default = get_target(local_cfg, 'default') or get_target(root_cfg, 'default')
            if custom_default is not None:
                local_target.set_parent(custom_default)
        local_target.resolve_options(problem)
        return local_target

    # Inheritance chain:
    # root_default -> [package_default] -> None
    # root_non_default -> [[root_default] -> package_default] -> None
    root_target = get_target(root_cfg, name)
    if root_target is not None:
        if root_target.need_default:
            root_target.set_parent(package_default)
            if name != 'default':
                root_default = get_target(root_cfg, 'default')
                if root_default is not None:
                    root_target.set_parent(root_default)
        root_target.resolve_options(problem)
        return root_target

    if name == 'default':
        package_target = package_default
    else:
        package_target = get_target(package_cfg, name)
        if package_target is None:
            # not found
            return None
        if package_target.need_default:
            package_target.set_parent(package_default)

    package_target.resolve_options(problem)
    return package_target



import os
import subprocess
import sys
from itertools import chain

import click

from kks.util.config import find_target


GPP_ARGS = [  # used only for test generator / solution
    'g++',
    '-std=gnu++17',
    '-g',
    '-O2',
    '-Werror',
    '-Wall',
    '-Wextra',
    '-ftrapv',
    '-pthread',
]

ASAN_ARGS = [
    '-fsanitize=address',
    '-fsanitize=undefined',
    '-fno-sanitize-recover=all',  # for RE in case of UB
]

ASAN_ENV = {
    'ASAN_OPTIONS': 'color=always',
}

VALGRIND_ARGS = [
    'valgrind',
    '--leak-check=full',
]


def compile_solution(directory, target_name, verbose, options):
    target = find_target(target_name)
    if target is None:
        click.secho(f'No target {target_name} found', fg='red', err=True)
        return None

    if verbose:
        click.secho(f'Selected target: {target}')

    if options.asan is None:
        options.asan = target.default_asan  # will be used in run_solution
    # gcc (clang) can compile c and asm files together, so everything should be ok
    source_files = list(chain(*[directory.glob(f) for f in target.files]))

    if len(source_files) == 0:
        click.secho('No source files found', fg='yellow', err=True)
        return None

    has_c = any(f.name.endswith('.c') for f in source_files)
    has_cpp = any(f.name.endswith('.cpp') for f in source_files)
    # NOTE asm + cpp compilation was not tested
    if has_c and has_cpp:
        click.secho('Cannot compile C and C++ together', fg='red', err=True)
        return None

    click.secho('Compiling... ', fg='green', err=True, nl=False)

    binary = _compile_solution(directory, source_files, target, verbose, options, has_cpp)

    if binary is None:
        click.secho('Compilation failed!', fg='red', err=True)
        return None

    binary_name = click.style(binary.relative_to(directory).as_posix(), fg='red', bold=True)
    click.secho(f'Successfully compiled binary {binary_name}', fg='green', err=True)

    return binary


def _compile_solution(workdir, files, target, verbose, options, cpp):
    compiler = target.cpp_compiler if cpp else target.compiler
    std = target.cpp_std if cpp else target.std
    compiler_args = [compiler, '-std='+std] + target.flags

    if not target.asm64bit and any(f.suffix.lower() == '.s' for f in files):
        compiler_args.append('-m32')

    return compile_gnu(
            workdir,
            files,
            options,
            compiler_args,
            linker_args=[f'-l{lib}' for lib in target.libs],
            out_file=target.out,
            verbose=verbose
    )


def compile_cpp(workdir, files, options, verbose=False):
    return compile_gnu(workdir, files, options, list(GPP_ARGS), verbose=verbose)


def compile_gnu(workdir, files, options, compiler_args, linker_args=[], out_file='', verbose=False):
    filenames = [path.absolute() for path in files]

    command = compiler_args
    if options.asan:
        command += ASAN_ARGS
    if out_file:
        command += ['-o', (workdir / out_file).absolute()]
    command += filenames
    command += linker_args

    if verbose:
        click.secho('\nExecuting "{}"'.format(' '.join(map(str, command))))

    p = subprocess.run(command, cwd=workdir)

    if p.returncode != 0:
        return None

    return workdir / (out_file or 'a.out')


def run_solution(binary, args, options, test_data, capture_output=True):
    args = [binary.absolute()] + args

    env = os.environ
    if options.asan:
        env = dict(env, **ASAN_ENV)

    if options.valgrind:
        args = VALGRIND_ARGS + args

    if test_data.is_file():
        with test_data.input_file.open('rb') as f_in:
            process = subprocess.run(args, env=env, stdin=f_in, capture_output=capture_output)
    elif test_data.is_data():
        process = subprocess.run(args, env=env, input=test_data.input_data, capture_output=capture_output)
    elif test_data.is_stdin():
        process = subprocess.run(args, env=env, stdin=sys.stdin, capture_output=capture_output)
    else:
        raise Exception('Unknown test type')

    return process




import click

from kks.errors import APIError
from kks.util.common import prompt_choice, with_retries
from kks.util.ejudge import RunStatus


class SubmissionResult:
    OK = 0
    CHECK = 1
    FAIL = 2
    UNKNOWN = 3

    _colors = {OK: 'green', FAIL: 'red', CHECK: 'bright_yellow', UNKNOWN: 'yellow'}

    def __init__(self, status, msg):
        self.status = status
        self.msg = msg

    def color(self):
        return self._colors[self.status]

    @classmethod
    def ok(cls, msg):
        return cls(cls.OK, msg)

    @classmethod
    def check(cls, msg):
        return cls(cls.CHECK, msg)

    @classmethod
    def fail(cls, msg):
        return cls(cls.FAIL, msg)

    @classmethod
    def unknown(cls, msg):
        return cls(cls.UNKNOWN, msg)

    @classmethod
    def parse_status(cls, run_status):
        if run_status.status in [RunStatus.OK, RunStatus.PENDING_REVIEW]:
            return cls.ok(str(run_status))
        if run_status.status in [RunStatus.ACCEPTED, RunStatus.PENDING]:  # PENDING == "Pending check" =?= ACCEPTED
            return cls.check(str(run_status))
        if run_status.status in [RunStatus.CE, RunStatus.STYLE_ERR]:
            return cls.fail(run_status.with_compiler_output())
        return cls.fail(run_status.with_tests(failed_only=True))  # there can be 100+ passed tests and a few failed


def get_lang(available, all_langs):
    def choice(langs):
        choices = [f"{lang['short_name']} - {lang['long_name']}" for lang in langs]
        lang_id = prompt_choice('Select a language / compiler', choices)
        return langs[lang_id]

    if not available:
        return None
    if len(available) == 1:
        return available[0]
    langs = [lang for lang in all_langs if lang['id'] in available]
    return choice(langs)['id']


def submit_solution(session, file, prob_name, timeout):

    @with_retries(step=2, timeout=timeout)
    def get_final_result(api, run_id):
        res = RunStatus(api.run_status(run_id))
        if res.is_testing():
            return None
        return SubmissionResult.parse_status(res)

    api = session.api()
    try:
        contest = session.with_auth(api.contest_status)
    except APIError as e:
        return SubmissionResult.fail(str(e))

    prob_id = None
    for p in contest['problems']:
        if p['short_name'] == prob_name:
            prob_id = p['id']
            break
    if prob_id is None:
        return SubmissionResult.fail('Invalid problem ID')

    problem = api.problem_status(prob_id)
    problem, problem_status = problem['problem'], problem['problem_status']
    if not problem_status.get('is_submittable'):
        return SubmissionResult.fail('Cannot submit a solution for this problem')
    ok_fields = ['is_solved', 'is_pending', 'is_pending_review', 'is_accepted']
    if any(field in problem_status for field in ok_fields) and not click.confirm('This problem was already solved! Submit anyway?'):
        return SubmissionResult.fail('Cancelled by user')

    lang = get_lang(problem.get('compilers', []), contest['compilers'])
    try:
        run_id = api.submit(prob_id, file, lang)['run_id']
    except APIError as e:  # Duplicate / empty file / etc.
        return SubmissionResult.fail(str(e))
    click.secho('Testing...', bold=True)
    return get_final_result(api, run_id) or SubmissionResult.unknown('Testing in progress')
